/*
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//
// @author Stefano Fornari
// @version $Id: TransportAgent.cpp,v 1.4 2004/02/09 08:48:50 ste Exp $
//

#include <winsock.h>

#include "common/Log.h"

#include "http/common/Constants.h"
#include "http/common/HTTPHeader.h"
#include "http/common/TransportAgent.h"


/*
 * This is the PocketPC implementation of the TransportAgent object
 */

/*
 * This function translate a UNICODE string into a UTF string without
 * allocating additional memory. The translation is performed removing
 * the second byte of the UNICODE coding.
 *
 * @param s the string to translate
 */
void toUTF(const wchar_t* s) {
    int l = wcslen(s);
    wchar_t* w = (wchar_t*)s;
    char*    c = (char*)s;

    while (l--) {
        *c = (char)*w;
        ++c; ++w;
    }

    *c = 0;
}



/*
 * Constructor.
 * In this implementation newProxy is ignored, since proxy configuration
 * is taken from the WinInet subsystem.
 *
 * @param url the url where messages will be sent with sendMessage()
 * @param proxy proxy information or NULL if no proxy should be used
 */
TransportAgent::TransportAgent(URL& newURL, Proxy& newProxy) {
    url = newURL;

    LOG.debug(TEXT("Initializing Winsock..."));
    if (init()) {
        LOG.debug(TEXT("Error initializing Winsock"));
        LOG.error(lastErrorMsg);
    }

    LOG.debug(TEXT("Winsock initialized"));
}

TransportAgent::~TransportAgent() {
    uninit();
}


/*
 * Change the URL the subsequent calls to setMessage() should
 * use as target url.
 *
 * @param url the new target url
 */
void TransportAgent::setURL(URL& newURL) {
    url = newURL;
}

/*
 * Sends the given SyncML message to the server specified
 * by the instal property 'url'. Returns the response status code or -1
 * if it was not possible initialize the connection.
 *
 * Use getResponse() to get the server response.
 */
wchar_t* TransportAgent::sendMessage(wchar_t* msg) {

    return postHTTP(url, msg);
}

// ------------------------------------------------------------- Private methods

int TransportAgent::init()
{
    WORD versionRequested = MAKEWORD( 1, 1 );
    WSADATA wsaData;
    int ret;

    //
    // Initialize WinSock.dll
    //
    // ret = WSAStartup(versionRequested, &wsaData); // originale
    ret = WSAStartup(versionRequested, (LPWSADATA)&wsaData);

    if (ret) {
        lastErrorCode = ret;
        wsprintf(lastErrorMsg, TEXT("WSAStartup() error (%d)"), ret);

        goto finally;
    }

    //
    // Check WinSock version
    //
    if (wsaData.wVersion != versionRequested) {
        ret = lastErrorCode = -1;
        wsprintf(lastErrorMsg, TEXT("WinSock version not supported"));

        goto finally;
    }

    LOG.debug(TEXT("3"));

finally:

    return ret;
}

void TransportAgent::uninit() {
    WSACleanup();
}


wchar_t* TransportAgent::postHTTP(URL& url, const wchar_t* msg) {
    LPHOSTENT   hostEntry;
    SOCKADDR_IN saServer;
    SOCKET      s;
    int         ret;
    char        host    [DIM_HOSTNAME];
    char        resource[DIM_URL     ];

    BOOL        usedDefaultChar       ;
    unsigned int l = wcslen(msg)      ;
    unsigned int n, read              ;
    unsigned int contentLength        ;
    wchar_t*    response = NULL       ;
    HTTPHeader* hdr = NULL            ;


    wsprintf(logmsg, TEXT("Getting %s://%s:%d%s"), url.protocol, url.host, url.port, url.resource);
    LOG.debug(logmsg);

    WideCharToMultiByte (CP_ACP, 0, url.host, DIM_HOSTNAME-1, host, DIM_HOSTNAME-1, "?", &usedDefaultChar);
    WideCharToMultiByte (CP_ACP, 0, url.resource, DIM_URL-1, resource, DIM_URL-1, "?", &usedDefaultChar);

    //
    // Lookup the host address
    //
    hostEntry = gethostbyname(host);

    if (hostEntry == NULL) {

        // try to retrieve host by address. Try to use AF_INET as protocol type. See definition in winsock.h
        u_long addr;
        addr = inet_addr(host);
        hostEntry = gethostbyaddr((char *) &addr, sizeof (addr), AF_INET);

        if (hostEntry == NULL) {

            lastErrorCode = ERR_CONNECT;

            wsprintf(lastErrorMsg, TEXT("Unknown host: %s"), url.host);

            goto finally;
        }
    }


    // Create a TCP/IP stream socket
    s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (s == INVALID_SOCKET) {
        lastErrorCode = ERR_CONNECT;
        wsprintf(lastErrorMsg, TEXT("Unable to create socket"));

        goto finally;
    }

    // Fill in the rest of the server address structure
    //
    saServer.sin_family = AF_INET;
    saServer.sin_addr = *((LPIN_ADDR)*hostEntry->h_addr_list);
    saServer.sin_port = htons(url.port);

    //
    // Connect the socket
    //
    ret = connect(s, (LPSOCKADDR)&saServer, sizeof(SOCKADDR_IN));
    if (ret == SOCKET_ERROR) {
        lastErrorCode = ERR_CONNECT;
        wsprintf(lastErrorMsg, TEXT("Unable to connect"));

        goto finally;
    }

    //
    // Format the HTTP request
    // and send it
    //
    char abuf[DIM_BUFFER];
    createHTTPRequest(abuf, resource, l);

    ret = send(s, abuf, strlen(abuf), 0);
    if (ret == SOCKET_ERROR) {
        lastErrorCode = ERR_CONNECT;
        wsprintf(lastErrorMsg, TEXT("Unknown sending data: %d"), ret);

        goto finally;
    }

    toUTF(msg);
    ret = send(s, (char*)msg, l, 0);
    if (ret == SOCKET_ERROR) {
        lastErrorCode = ERR_CONNECT;
        wsprintf(lastErrorMsg, TEXT("Unknown sending data: %d"), ret);

        goto finally;
    }

    //
    // We're connected, so we can now receive the content. When we read the first
    // buffer of data, we check the HTTP response headers looking for the status
    // and the content length.
    //

    read = recv(s, abuf, sizeof(abuf)-1, 0);
    if (read == SOCKET_ERROR) {
        lastErrorCode = ERR_READING_CONTENT;
        wsprintf(lastErrorMsg, TEXT("Error freading content: %d"), WSAGetLastError());

        goto finally;
    }
    abuf[read] = 0;

    hdr = new HTTPHeader(abuf);

    if (hdr->getStatus() != STATUS_OK) {
        char* msg = hdr->getStatusMessage();

        lastErrorCode = hdr->getStatus();

        if (msg != NULL) {
            int l = strlen(msg);

            MultiByteToWideChar (CP_ACP, 0, msg, l, lastErrorMsg, l);
        } else {
            wsprintf(lastErrorMsg, TEXT("HTTP error %s"), lastErrorCode);
        }

        goto finally;
    }

    contentLength = (unsigned)hdr->getContentLength();

    if (((int)contentLength) <= 0) {
        lastErrorCode = ERR_HTTP_MISSING_CONTENT_LENGTH;
        wsprintf(lastErrorMsg, TEXT("Content length missing or zero"));

        goto finally;
    }

    response = new wchar_t[contentLength+1];

    n = min(contentLength, DIM_BUFFER - (hdr->getSize()));

    MultiByteToWideChar (CP_ACP, 0, hdr->getContent(), n, response, n);

    //
    // Receive the file contents and print to stdout
    //
    read -= hdr->getSize();
    if (read > contentLength) {
        read = contentLength;
    }
    while(read < contentLength) {
        // Wait to receive, nRet = NumberOfBytesReceived
        ret = recv(s, abuf, sizeof(abuf)-1, 0);
        if (ret == SOCKET_ERROR) {
            lastErrorCode = ERR_CONNECT;
            wsprintf(lastErrorMsg, TEXT("Error freading data: %d"), WSAGetLastError());

            goto finally;
        }

        abuf[ret] = 0;

        wsprintf(logmsg, TEXT("recv() returned %d bytes"), ret);
        LOG.debug(logmsg);

        // Did the server close the connection?
        if (ret == 0) {
            break;
        }

        if (read + ret > contentLength) {
            n = contentLength - read;
        } else {
            n = ret;
        }

        // Write to log

        MultiByteToWideChar (CP_ACP, 0, abuf, n, &response[read], n);

        read += n;
    }

    //
    // If the number of read bytes is different from the content length
    // something was wrong...
    //
    response[min(read, contentLength)] = 0;
    if (read != contentLength) {
        wsprintf(logmsg, TEXT("Bytes read (%u) and Content-Length (%u) do not match!"), read, hdr->getContentLength());
        LOG.info(logmsg);
    }

finally:

    if (hdr != NULL) {
        delete hdr;
    }

    if (s != NULL) {
        closesocket(s);
    }

    return response;
}

/*
 * Create the POST HTTP request given the resource and the length of
 * data to be posted (as number of bytes).
 *
 * @param reuqest  - OUT - the buffer where the request is written
 * @param resource - IN  - the resource to address in the request
 * @param length   - IN  - content-length
 */
TransportAgent::createHTTPRequest( char*        request  ,
                                   const char*  resource ,
                                   unsigned int length   ) {
    sprintf(
        request,
        "POST %s HTTP/1.0\nContent-Type: %s\nContent-Length: %d\nUser-Agent: %s\n\n",
        resource,
        SYNCML_CONTENT_TYPE,
        length,
        USER_AGENT
    );


}




