/*
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//
// @author Stefano Fornari
// @version $Id: SyncMLProcessor.cpp,v 1.6 2004/02/09 08:48:50 ste Exp $
//

#include <stdlib.h>
#include <stdio.h> // to be removed

#include "spds/common/Utils.h"
#include "spds/common/Constants.h"
#include "spds/common/SyncMLProcessor.h"
#include "spds/common/SyncItemHolder.h"

/*
 * This class is responsible for the processing of the incoming messages.
 */


USING_PTYPES

/*
 * Constructor
 */
SyncMLProcessor::SyncMLProcessor() {
}

/*
 * Processes the initialization response. Returns 0 in case of success, an
 * error code in case of error.
 *
 * @param msg the response from the server
 */
int SyncMLProcessor::processInitResponse(SyncSource& source, wchar_t* msg) {
    wchar_t* syncBody = NULL;
    wchar_t* alert    = NULL;
    wchar_t* item     = NULL;
    wchar_t* target   = NULL;
    wchar_t* uri      = NULL;
    wchar_t* data     = NULL;
    unsigned int n, pos;
    int ret = -1;


    //
    // We have to look for the Alert element relative to the given source.
    // We will store the alert code in the SyncSource object.
    // If no valid Alert is found, we set lastErrorCode and lastErrorMsg
    // accordingly.
    //
    syncBody = getElementContent(msg, TEXT("SyncBody"), NULL);
    if (syncBody == NULL) {
        ret = lastErrorCode = ERR_REPRESENTATION;
        wsprintf(lastErrorMsg, TEXT("SyncBody not found!"));

        goto finally;
    }

    //
    // First of all check the status for the SyncHead
    //
    ret = getSyncHeaderStatusCode(syncBody);
    if ((ret <= 200) || (ret >=299)) {
        goto finally;
    }

    //
    // If the access is allowed
    ret = getAlertStatusCode(syncBody, source.getName(NULL, 0));
    if ((ret < 200) || (ret > 299)) {
        goto finally;
    }

    ret = 0;

    pos = n = 0;
    do {
        alert = getElementContent(syncBody, TEXT("Alert"), &n);
        if (alert == NULL) {
            break; // exit loop
        }
        pos += n;

        item = getElementContent(alert, TEXT("Item"), NULL);
        if (item == NULL) {
            lastErrorCode = ERR_REPRESENTATION;
            wsprintf(lastErrorMsg, TEXT("SyncBody/Alert/Item not found!"));

            goto finally;
        }

        target = getElementContent(item, TEXT("Target"), NULL);
        if (item == NULL) {
            lastErrorCode = ERR_REPRESENTATION;
            wsprintf(lastErrorMsg, TEXT("SyncBody/Alert/Item/Target not found!"));

            goto finally;
        }

        uri = getElementContent(target, TEXT("LocURI"), NULL);
        if (uri == NULL) {
            lastErrorCode = ERR_REPRESENTATION;
            wsprintf(lastErrorMsg, TEXT("SyncBody/Alert/Item/Target/LocURI not found!"));

            goto finally;
        }

        if (wcscmp(uri, source.getName(NULL, 0)) == 0) {
            data = getElementContent(alert, TEXT("Data"), NULL);
            if (data == NULL) {
                lastErrorCode = ERR_REPRESENTATION;
                wsprintf(lastErrorMsg, TEXT("SyncBody/Alert/Data not found!"));

                goto finally;
            }

            source.setSyncMode((SyncMode)wcstol(data, NULL, 10));
            ret = 0;
            break;
        }

        //
        // For the next iteration we will have new values
        //
        safeDelete(&syncBody);
        safeDelete(&alert   );
        safeDelete(&item    );
        safeDelete(&target  );
        safeDelete(&uri     );
        safeDelete(&data    );
    } while (alert != NULL);

    ret = 0;

    //
    // Finally ...
    //
finally:

    safeDelete(&syncBody);
    safeDelete(&alert   );
    safeDelete(&item    );
    safeDelete(&target  );
    safeDelete(&uri     );
    safeDelete(&data    );

    return ret;
}

/*
 * Processes the synchronization response. Returns 0 in case of success, an
 * error code in case of error.
 * It feeds the given source with the server side modifications
 *
 * @param source the source
 * @param msg the response from the server
 */
int SyncMLProcessor::processSyncResponse(SyncSource& source, wchar_t* msg) {
    unsigned int n = 0;
    unsigned int pos1 = 0, pos2 = 0, pos3 = 0;
    unsigned int syncCount = 0, modCount = 0, itemCount = 0;
    int ret = -1;

    wchar_t* sync   = NULL;
    wchar_t* mod    = NULL;
    wchar_t* target = NULL;
    wchar_t* uri    = NULL;
    wchar_t* item   = NULL;
    wchar_t* data   = NULL;
    wchar_t* id     = NULL;

    SyncItem* syncItem = NULL;

    tobjlist<SyncItem*> items = tobjlist<SyncItem*>();

    //
    // First of all check the status for the SyncHead
    //
    ret = getSyncHeaderStatusCode(msg);
    if ((ret < 200) || (ret >299)) {
        goto finally;
    }

    do {
        ++syncCount; pos2 = pos3 = 0;
        sync = getElementContent(&msg[pos1], TEXT("Sync"), &n);
        if (sync == NULL) {
            //
            // No more to do
            //
            break;
        }
        pos1 += n;

        target = getElementContent(sync, TEXT("Target"), NULL);
        if (target == NULL) {
            //
            // It should not happen
            //
            ret = lastErrorCode = ERR_REPRESENTATION;
            wsprintf(lastErrorMsg, TEXT("Missing Sync[%d]/Target element!"), syncCount);

            goto finally;
        }

        uri = getElementContent(target, TEXT("LocURI"), NULL);
        if ((uri == NULL) || (wcscmp(uri, source.getName(NULL, 0)) != 0)) {
            //
            // This Sync is for another source
            //
            safeDelete(&target);
            safeDelete(&uri   );
            continue;
        }

        //
        // Here the Sync is for the given source
        //
        safeDelete(&target);
        safeDelete(&uri   );

        //
        // Search for all Add
        //
        clear(items); modCount = 0;
        do {
            ++modCount; pos3 = 0;

            safeDelete(&item);
            safeDelete(&id  );
            safeDelete(&data);
            safeDelete(&mod );

            mod = getElementContent(&sync[pos2], TEXT("Add"), &n);
            if (mod == NULL) {
                //
                // No more Add
                //
                break;
            }
            pos2 += n;

            //
            // Search for all Item
            //
            itemCount = 0;
            do {
                ++itemCount;

                safeDelete(&item);
                safeDelete(&id  );
                safeDelete(&data);

                item = getElementContent(&mod[pos3], TEXT("Item"), &n);
                if (item == NULL) {
                    //
                    // No more Item
                    //
                    break;
                }
                pos3 += n;

                id = getElementContent(item, TEXT("LocURI"), NULL);
                if (id == NULL) {
                    //
                    // Representation error
                    //
                    ret = lastErrorCode = ERR_REPRESENTATION;
                    wsprintf(lastErrorMsg, TEXT("Missing Sync[%d]/Add[%d]/Item[i]/Source/LocURI element!"), syncCount, modCount, itemCount);

                    goto finally;
                }

                data = getElementContent(item, TEXT("Data"), NULL);

                syncItem = new SyncItem(id);
                syncItem->setData(data, (data == NULL) ? 0 : (wcslen(data)+1)*sizeof(wchar_t));

                add(items, new SyncItemHolder(syncItem));
            } while (item != NULL);
        } while (mod != NULL);

        source.setNewSyncItems(toSyncItemArray(items), length(items));

        //
        // Search for all Replace
        //
        clear(items); modCount = 0; pos2 = 0;
        do {
            ++modCount; pos3 = 0;

            safeDelete(&item);
            safeDelete(&id  );
            safeDelete(&data);
            safeDelete(&mod);

            mod = getElementContent(&sync[pos2], TEXT("Replace"), &n);
            if (mod == NULL) {
                //
                // No more Replace
                //
                break;
            }
            pos2 += n;

            //
            // Search for all Item
            //
            itemCount = 0;
            do {

                ++itemCount;

                safeDelete(&item);
                safeDelete(&id  );
                safeDelete(&data);

                item = getElementContent(&mod[pos3], TEXT("Item"), &n);
                if (item == NULL) {
                    //
                    // No more Item
                    //
                    break;
                }
                pos3 += n;

                id = getElementContent(item, TEXT("LocURI"), NULL);
                if (id == NULL) {
                    //
                    // Representation error
                    //
                    ret = lastErrorCode = ERR_REPRESENTATION;
                    wsprintf(lastErrorMsg, TEXT("Missing Sync[%d]/Replace[%d]/Item[i]/Source/LocURI element!"), syncCount, modCount, itemCount);

                    goto finally;
                }

                data = getElementContent(item, TEXT("Data"), NULL);

                syncItem = new SyncItem(id);
                syncItem->setData(data, (data == NULL) ? 0 : (wcslen(data)+1)*sizeof(wchar_t));

                add(items, new SyncItemHolder(syncItem));
            } while (item != NULL);
        } while (mod != NULL);

        source.setUpdatedSyncItems(toSyncItemArray(items), length(items));

        //
        // Search for all Delete
        //
        clear(items); modCount = 0; pos2 = 0;
        do {
            ++modCount; pos3 = 0;

            safeDelete(&item);
            safeDelete(&id  );
            safeDelete(&data);
            safeDelete(&mod);


            mod = getElementContent(&sync[pos2], TEXT("Delete"), &n);
            if (mod == NULL) {
                //
                // No more Delete
                //
                break;
            }
            pos2 += n;

            //
            // Search for all Item
            //
            itemCount = 0;
            do {
                ++itemCount;

                safeDelete(&item);
                safeDelete(&id  );
                safeDelete(&data);

                item = getElementContent(&mod[pos3], TEXT("Item"), &n);
                if (item == NULL) {
                    //
                    // No more Item
                    //
                    break;
                }
                pos3 += n;

                id = getElementContent(item, TEXT("LocURI"), NULL);
                if (id == NULL) {
                    //
                    // Representation error
                    //
                    ret = lastErrorCode = ERR_REPRESENTATION;
                    wsprintf(lastErrorMsg, TEXT("Missing Sync[%d]/Delete[%d]/Item[i]/Source/LocURI element!"), syncCount, modCount, itemCount);

                    goto finally;
                }

                syncItem = new SyncItem(id);
                syncItem->setData(NULL, 0);

                add(items, new SyncItemHolder(syncItem));
            } while (item != NULL);
        } while (mod != NULL);

        source.setDeletedSyncItems(toSyncItemArray(items), length(items));

    } while (sync != NULL);

    clear(items);  // releases the SyncItemHolders

    ret = 0;

finally:

    safeDelete( &sync   );
    safeDelete( &mod    );
    safeDelete( &target );
    safeDelete( &uri    );
    safeDelete( &item   );
    safeDelete( &data   );
    safeDelete( &id    );

    return ret;

}

/*
 * Processes the map message response. Returns 0 in case of success, an
 * error code in case of error.
 * It feeds the given source with the server side modifications
 *
 * @param source the source
 * @param msg the response from the server
 */
int SyncMLProcessor::processMapResponse(SyncSource& source, wchar_t* msg) {
    int ret = -1;

    //
    // for now it is always ok
    //
    //
    // First of all check the status for the SyncHead
    //
    ret = getSyncHeaderStatusCode(msg);
    if ((ret < 200) || (ret >299)) {
        goto finally;
    }

    ret = 0;

finally:

    return ret;
}

/*
 * Returns the SyncHeader/RespURI element of the given message. If the element is not
 * found it returns NULL. The returned respURI is allocated with the new operator
 * and must be discarded with delete by the caller.
 *
 * @param msg - the message - NOT NULL
 */
wchar_t* SyncMLProcessor::getRespURI(wchar_t* msg) {
    wchar_t* header  = NULL;
    wchar_t* respURI = NULL;

    header = getElementContent(msg, TEXT("SyncHdr"), NULL);
    if (header == NULL) {
        goto finally;
    }

    respURI = getElementContent(header, TEXT("RespURI"), NULL);

    finally:

    safeDelete(&header);

    return respURI;
}


// ------------------------------------------------------------- Private methods

/*
 * Extracts the content of a tag into a SyncML message. It is supposed the
 * message is a valid SyncML message. It returns NULL in case the tag is not
 * found or the XML fragment is not in the expected form.
 * The returned pointer (if not NULL) is allocated with the new operator and
 * must be discarded with the operator delete.
 *
 * @param xml the xml fragment
 * @param tag the tag we want the content
 * @param pos (OUTPUT) the position just after the closing tag (ignored if NULL)
 *
 */
wchar_t* SyncMLProcessor::getElementContent(wchar_t* xml, wchar_t* tag, unsigned int* pos) {
    wchar_t* p1       = NULL;
    wchar_t* p2       = NULL;
    wchar_t* ret      = NULL;

    wchar_t openTag[16];
    wchar_t closeTag[16];

    unsigned int l = wcslen(tag);

    if (pos != NULL) {
        *pos = 0;
    }

    wsprintf(openTag, TEXT("<%s>"), tag);
    wsprintf(closeTag, TEXT("</%s>"), tag);

    p1 = wcsstr(xml, openTag);

    if (p1 == NULL) {
        //
        // This is abcxyz
        //
        goto finally;
    }
    p1 = p1+l+2;

    if (*p1 == 0) {
        //
        // This is abc<tag>\0
        //
        goto finally;
    }

    p2 = wcsstr(p1, closeTag);

    if (p2 == NULL) {
        //
        // This is abc<tag>xyz\0
        //
        goto finally;
    }

    ret = new wchar_t[p2-p1+1];

    wcsncpy(ret, p1, p2-p1);
    ret[p2-p1] = 0;

    if (pos != NULL) {
        *pos = p2-xml+l+3;
    }

    finally:

    return ret;

}

/*
 * Returns the status code for the SyncHeader command included
 * in the message sent by the client.
 *
 * @param syncBody - the SyncBody content
 */
int SyncMLProcessor::getSyncHeaderStatusCode(wchar_t* syncBody) {
    wchar_t* status = NULL;
    wchar_t* cmdRef = NULL;
    wchar_t* data   = NULL;

    int ret = -1;

    //
    // It must be the first one
    //
    status = getElementContent(syncBody, TEXT("Status"), NULL);
    if (status == NULL) {
        //
        // It should not happen
        //
        wsprintf(lastErrorMsg, TEXT("Status command for SyncHeader not found!"));
        lastErrorCode = ERR_REPRESENTATION;
        goto finally;
    }

    cmdRef = getElementContent(status, TEXT("CmdRef"), NULL);
    if ((cmdRef == NULL) || (wcscmp(cmdRef, TEXT("0")))) {
        //
        // It should not happen
        //
        wsprintf(lastErrorMsg, TEXT("Status/CmdRef or not found or not referring to SyncHeader!"));
        lastErrorCode = ERR_REPRESENTATION;
        goto finally;
    }

    data = getElementContent(status, TEXT("Data"), NULL);
    if (data == NULL) {
        //
        // It should not happen
        //
        wsprintf(lastErrorMsg, TEXT("Status/Data not found!"));
        lastErrorCode = ERR_REPRESENTATION;
        goto finally;
    }

    ret = wcstoul(data, NULL, 10);

    finally:

    safeDelete(&status);
    safeDelete(&cmdRef);
    safeDelete(&data  );

    return ret;
}


/*
 * Returns the status code for the Alert relative to the given source.
 *
 * @param syncBody - the SyncBody content
 * @param sourceName - the name of the source
 */
int SyncMLProcessor::getAlertStatusCode(wchar_t* syncBody, wchar_t* sourceName) {
    unsigned int pos = 0;

    wchar_t* status    = NULL;
    wchar_t* cmd       = NULL;
    wchar_t* data      = NULL;
    wchar_t* sourceRef = NULL;

    int ret = -1;

    do {
        safeDelete(&status   );
        safeDelete(&cmd      );
        safeDelete(&sourceRef);

        status = getElementContent(&syncBody[pos], TEXT("Status"), &pos);
        if (status == NULL) {
            break;
        }

        cmd = getElementContent(status, TEXT("Cmd"), NULL);
        if (cmd == NULL) {
            //
            // It should not happen
            //
            wsprintf(lastErrorMsg, TEXT("Status/Cmd not found!"));
            lastErrorCode = ERR_REPRESENTATION;
            goto finally;
        }

        if (wcscmp(TEXT("Alert"), cmd) != 0) {
            continue;
        }

        sourceRef = getElementContent(status, TEXT("SourceRef"), NULL);
        if ((sourceRef != NULL) && (wcscmp(sourceRef, sourceName) == 0)) {
            //
            // Found!
            //
            break;
        }

    } while(status != NULL);

    //
    // If status is not null, it is the good one!
    //
    if (status != NULL) {
        data = getElementContent(status, TEXT("Data"), NULL);
        if (data == NULL) {
            //
            // It should not happen
            //
            wsprintf(lastErrorMsg, TEXT("Status/Data not found!"));
            lastErrorCode = ERR_REPRESENTATION;
            goto finally;
        }
        ret = wcstoul(data, NULL, 10);
    }

    finally:

    safeDelete(&status   );
    safeDelete(&cmd      );
    safeDelete(&data     );
    safeDelete(&sourceRef);

    return ret;
}

/*
 * Trnslates a PTypes object list into an array of SyncItem*.
 *
 * @param items the item list
 */
SyncItem** SyncMLProcessor::toSyncItemArray(tobjlist<SyncItem*>& items) {
    int l = length(items);

    SyncItem** itemArray = new SyncItem*[l];

    for (int i=0; i<l; ++i) {
        itemArray[i] = ((SyncItemHolder*)items[i])->getItem();
    }

    return itemArray;
}

